/* bf.c
******************************************************************************

   Yet another Brainfuck interpreter

   Author: Stephan Beyer
   
   Copyright (C) GPL, 2003 Stephan Beyer - s-beyer@gmx.net
  
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h> /* malloc, atoi */
#include <string.h> /* strlen */
#include "errors.h"

#define VERSION "20031030"
#define CELLS 9999

#define Debug(X)	if (opt.debug) fputs("DEBUG: "#X"\n", stderr);
#define ShowInput(X)	if (opt.showinput) fprintf(stderr, "%c", c);

/* Brainfuck array and pointer */
char *array;
char *p;

/* options */
typedef struct {
	int cells;
	char *filename;
	unsigned debug : 1;
	unsigned showinput : 1;
	unsigned tellarray : 1; /* when someone needs this feature it will
				   be implemented */
} Opt;
Opt opt;

/* begins interpretion in filehandle fp at <start> end returns where it
 * ended (either at ] or eof - on eof it returns -1)  */
long Interprete (FILE *fp, long start)
{
	unsigned char c;
	int do_nothing = 0;

	DoAndErr(fseek(fp, start, SEEK_SET), fseek, errFILE);

	while (!feof(fp))
	{
		c = (unsigned char)fgetc(fp);

		if ((do_nothing) && (c != ']')) {
			if (c == '[')
				do_nothing++;
			continue;
		}

		switch(c)
		{
			case '>':
				if (p+1 > array+opt.cells)
					ErrorMsg (errOUTOFRANGE, "Out of range! You wanted to '>' beyond the last cell. See -c option.");
				++p;
				ShowInput(c);
				break;
			case '<':
				if (p-1 < array)
					ErrorMsg (errOUTOFRANGE, "Out of range! You wanted to '<' below the first cell.");
				--p;
				ShowInput(c);
				break;
			case '+':
				++*p;
				ShowInput(c);
				break;
			case '-':
				--*p;
				ShowInput(c);
				break;
			case '.':
				ShowInput(c);
				putchar(*p);
				fflush(stdout);
				break;
			case ',':
				ShowInput(c);
				*p = getchar();
				break;
			case '[':
			{
				long pos = ftell(fp);
				ShowInput(c);

				if (!*p)
				{
					do_nothing++;
					Debug(skip loop)
					break;
				}
				Debug(begin loop);

				while(*p) {
					Debug(loop)
					Interprete(fp, pos);
				}

				Debug(loop ended);

				/* we should still be at the position where
				 * we ought to be, so I won't fseek */
			}	break;
			case ']':
				if (do_nothing)
					do_nothing--;
				else {
					ShowInput(c);
					return ftell(fp);
				}
		}
	}
	return -1;
}

void ResetArray()
{
	int i;
	array = malloc(opt.cells+2);
	DoAndErr(!array, malloc, errMEMORY);
	p = array;

	for (i = 0; i < opt.cells; i++)
		*(p++) = 0;

	p = array;
}

/* display usage information */
void Usage(char *bin)
{
	printf	("Yet another Brainfuck interpreter       version %s\n", VERSION);
	puts	("(C) 2003, Stephan Beyer, GPL, s-beyer@gmx.net\n");
	puts	("Usage information: ");
	printf	("\t%s [options] inputfile\n\n", bin);
	puts	("Available options:");
	printf	("\t-c<num>   specify number of cells [%d]\n", CELLS);
	puts	("\t-d        show some debugging information (stderr)");
	puts	("\t-i        show used input (stderr)");
	puts	("\nHave fun!");
}

void HandleOptions (int c, char **v)
/* too lazy to use getopt this time ;) */
{
	int i;
	opt.showinput = 0;
	opt.cells = CELLS;
	opt.debug = 0;
	
	if (c <= 1)
	{
		Usage(*v);
		ErrorMsg (errOPT, "No input file given.");
	}
	
	for(i = 1; i < c-1; i++)
	{
		if ((*v[i] == '-')
		  &&(strlen(v[i]) >= 2))
		{
			switch(*(v[i]+1))
			{
				case 'c':
					opt.cells = atoi(v[i]+2);
					break;
				case 'd':
					opt.debug = 1;
					break;
				case 'i':
					opt.showinput = 1;
					break;
				default:
					Usage(*v);
					ErrorMsg(errOPT, "Unknown option.");
			}
		} else {
			Usage(*v);
			ErrorMsg(errOPT, "Invalid option(s).");
		}
	}
	opt.filename = v[i]; /* last option MUST be filename */
}

int main (int argc, char **argv)
{
	FILE *fp;

	/* options */
	HandleOptions(argc, argv);

	ResetArray();

	/* file opening banana */
	fp = fopen(opt.filename, "r");
	DoAndErr(!fp, fopen, errFILE); 


	/* reading and interpreting */
	if(Interprete(fp, 0) > -1)
		ErrorMsg(errBRACKET, "Unmatching brackets!");

	/* cleanup */
	fclose(fp);
	free(array);
	return 0;
}
